@@ -49,12 +49,20 @@
 #include "cafebabe/method_info.h"
 #include "cafebabe/stream.h"
 
+#include "runtime/stack-walker.h"
+#include "runtime/runtime.h"
+#include "runtime/unsafe.h"
+#include "runtime/class.h"
+#include "runtime/classloader.h"
+
 #include "jit/compiler.h"
 #include "jit/cu-mapping.h"
+#include "jit/gdb.h"
 #include "jit/exception.h"
 #include "jit/perf-map.h"
 #include "jit/text.h"
 
+#include "lib/parse.h"
 #include "lib/list.h"
 
 #include "vm/fault-inject.h"
@@ -72,7 +80,6 @@
 #include "vm/string.h"
 #include "vm/system.h"
 #include "vm/thread.h"
-#include "vm/unsafe.h"
 #include "vm/class.h"
 #include "vm/call.h"
 #include "vm/jar.h"
@@ -82,6 +89,7 @@
 
 #include "arch/init.h"
 
+static bool dump_maps;
 static bool perf_enabled;
 static char *exe_name;
 
@@ -104,62 +112,6 @@ static void __attribute__((noreturn)) vm_exit(int status)
 	error("System.exit() returned");
 }
 
-static struct vm_object *native_vmstackwalker_getclasscontext(void)
-{
-	struct stack_trace_elem st_elem;
-	struct compilation_unit *cu;
-	struct vm_object *res;
-	int nr_to_skip;
-	int depth;
-	int idx;
-	int i;
-
-	init_stack_trace_elem_current(&st_elem);
-	depth = get_java_stack_trace_depth(&st_elem);
-
-	/* Skip call to VMStackWalker.getClassContext() */
-	depth -= 2;
-	stack_trace_elem_next_java(&st_elem);
-	stack_trace_elem_next_java(&st_elem);
-
-	res = vm_object_alloc_array(vm_array_of_java_lang_Class, depth);
-	if (!res) {
-		signal_new_exception(vm_java_lang_OutOfMemoryError, NULL);
-		return NULL;
-	}
-
-	nr_to_skip = 0;
-	idx = 0;
-
-	for (i = 0; i < depth; i++, stack_trace_elem_next_java(&st_elem)) {
-		cu = jit_lookup_cu(st_elem.addr);
-		if (!cu)
-			error("no compilation_unit mapping for %lx", st_elem.addr);
-
-		struct vm_class *vmc = cu->method->class;
-
-		if (vmc == vm_java_lang_reflect_Method &&
-		    !strcmp(cu->method->name, "invoke")) {
-			/*
-			 * We should skip all Method.invoke() frames
-			 * here as stated in comment for
-			 * VMStackWalker.getClassContext().
-			 */
-			nr_to_skip++;
-			continue;
-		}
-
-		if (vm_class_ensure_init(vmc))
-			return NULL;
-
-		array_set_field_object(res, idx++, vmc->object);
-	}
-
-	res->array_length -= nr_to_skip;
-
-	return res;
-}
-
 static void vm_properties_set_property(struct vm_object *p,
 				       const char *key, const char *value)
 {
@@ -263,13 +215,13 @@ static struct system_property system_properties[] = {
 	{ "file.separator",			"/"				},
 	{ "path.separator",			":"				},
 	{ "line.separator",			"\n"				},
-	{ "java.version",			"1.4"				},
+	{ "java.version",			"1.6"				},
 	{ "java.vendor",			"GNU Classpath"			},
 	{ "java.vendor.url",			"http://www.classpath.org"	},
 	{ "java.vm.specification.version",	"1.0"				},
 	{ "java.vm.specification.vendor",	"Sun Microsystems, Inc."	},
 	{ "java.vm.specification.name",		"Java Virtual Machine Specification"},
-	{ "java.specification.version",		"1.4"				},
+	{ "java.specification.version",		"1.6"				},
 	{ "java.specification.vendor",		"Sun Microsystems, Inc."	},
 	{ "java.specification.name",		"Java Platform API Specification"},
 	{ "java.class.version",			"48.0"				},
@@ -341,75 +293,6 @@ static void native_vmsystemproperties_preinit(struct vm_object *p)
 	}
 }
 
-static void native_vmruntime_exit(int status)
-{
-	/* XXX: exit gracefully */
-	exit(status);
-}
-
-static void native_vmruntime_run_finalization_for_exit(void)
-{
-}
-
-static struct vm_object *
-native_vmruntime_maplibraryname(struct vm_object *name)
-{
-	struct vm_object *result;
-	char *str;
-
-	if (!name) {
-		signal_new_exception(vm_java_lang_NullPointerException, NULL);
-		return NULL;
-	}
-
-	str = vm_string_to_cstr(name);
-	if (!str) {
-		NOT_IMPLEMENTED;
-		return NULL;
-	}
-
-	char *result_str = NULL;
-
-	if (asprintf(&result_str, "lib%s.so", str) < 0)
-		die("asprintf");
-
-	free(str);
-
-	if (!result_str) {
-		NOT_IMPLEMENTED;
-		return NULL;
-	}
-
-	result = vm_object_alloc_string_from_c(result_str);
-	free(result_str);
-
-	return result;
-}
-
-static int
-native_vmruntime_native_load(struct vm_object *name,
-			     struct vm_object *classloader)
-{
-	char *name_str;
-	int result;
-
-	if (!name) {
-		signal_new_exception(vm_java_lang_NullPointerException, NULL);
-		return 0;
-	}
-
-	name_str = vm_string_to_cstr(name);
-	if (!name_str) {
-		NOT_IMPLEMENTED;
-		return 0;
-	}
-
-	result  = vm_jni_load_object(name_str, classloader);
-	free(name_str);
-
-	return result == 0;
-}
-
 static void native_vmruntime_println(struct vm_object *message)
 {
 	if (!message) {
@@ -466,7 +349,7 @@ native_vmsystem_arraycopy(struct vm_object *src, int src_start,
 		return;
 	}
 
-	elem_size = get_vmtype_size(elem_type);
+	elem_size = vmtype_get_size(elem_type);
 	memmove(dest->fields + dest_start * elem_size,
 		src->fields + src_start * elem_size,
 		len * elem_size);
@@ -521,170 +404,6 @@ native_vmobject_getclass(struct vm_object *object)
 	return object->class->object;
 }
 
-static struct vm_object *
-native_vmclass_getclassloader(struct vm_object *object)
-{
-	struct vm_class *vmc;
-
-	vmc = vm_object_to_vm_class(object);
-	if (!vmc)
-		return NULL;
-
-	return vmc->classloader;
-}
-
-static struct vm_object *
-native_vmclass_forname(struct vm_object *name, jboolean initialize,
-		       struct vm_object *loader)
-{
-	struct vm_class *class;
-	char *class_name;
-
-	if (!name) {
-		signal_new_exception(vm_java_lang_NullPointerException, NULL);
-		return NULL;
-	}
-
-	class_name = vm_string_to_cstr(name);
-	if (!class_name) {
-		signal_new_exception(vm_java_lang_OutOfMemoryError, NULL);
-		return NULL;
-	}
-
-	class = classloader_load(loader, class_name);
-	if (!class)
-		goto throw_cnf;
-
-	if (initialize) {
-		if (vm_class_ensure_init(class))
-			goto throw_cnf;
-	}
-
-	return class->object;
- throw_cnf:
-	signal_new_exception(vm_java_lang_ClassNotFoundException, class_name);
-	free(class_name);
-	return NULL;
-}
-
-static struct vm_object *
-native_vmclass_getname(struct vm_object *object)
-{
-	struct vm_class *class = vm_object_to_vm_class(object);
-
-	if (!class)
-		return NULL;
-
-	struct vm_object *obj;
-	char *dot_name;
-
-	dot_name = slash2dots(class->name);
-	obj =  vm_object_alloc_string_from_c(dot_name);
-	free(dot_name);
-
-	return obj;
-}
-
-static int32_t
-native_vmclass_is_anonymous_class(struct vm_object *object)
-{
-	struct vm_class *class = vm_object_to_vm_class(object);
-
-	if (!class)
-		return false;
-
-	return vm_class_is_anonymous(class);
-}
-
-static jboolean
-native_vmclass_is_assignable_from(struct vm_object *clazz_1, struct vm_object *clazz_2)
-{
-	struct vm_class *vmc_1 = vm_object_to_vm_class(clazz_1);
-	struct vm_class *vmc_2 = vm_object_to_vm_class(clazz_2);
-
-	if (!vmc_1 || !vmc_2)
-		return false;
-
-	return vm_class_is_assignable_from(vmc_1, vmc_2);
-}
-
-static int32_t
-native_vmclass_isarray(struct vm_object *object)
-{
-	struct vm_class *class = vm_object_to_vm_class(object);
-
-	if (!class)
-		return false;
-
-	return vm_class_is_array_class(class);
-}
-
-static int32_t
-native_vmclass_isprimitive(struct vm_object *object)
-{
-	struct vm_class *class = vm_object_to_vm_class(object);
-
-	if (!class)
-		return false;
-
-	return vm_class_is_primitive_class(class);
-}
-
-static jint native_vmclass_getmodifiers(struct vm_object *clazz)
-{
-	struct vm_class *class = vm_object_to_vm_class(clazz);
-	if (!class)
-		return 0;
-
-	return class->access_flags;
-}
-
-static struct vm_object *
-native_vmclass_getcomponenttype(struct vm_object *object)
-{
-	struct vm_class *class = vm_object_to_vm_class(object);
-
-	if (!class)
-		return NULL;
-
-	if (!vm_class_is_array_class(class)) {
-		warn("%s", class->name);
-		return NULL;
-	}
-
-	return vm_class_get_array_element_class(class)->object;
-}
-
-static struct vm_object *
-native_vmclassloader_getprimitiveclass(int type)
-{
-	char primitive_class_name[] = { "X" };
-	struct vm_class *class;
-
-	primitive_class_name[0] = (char)type;
-
-	class = classloader_load_primitive(primitive_class_name);
-	if (!class)
-		return NULL;
-
-	vm_class_ensure_init(class);
-	if (exception_occurred())
-		return NULL;
-
-	return class->object;
-}
-
-static jboolean
-native_vmclass_isinstance(struct vm_object *clazz, struct vm_object *object)
-{
-	struct vm_class *class = vm_object_to_vm_class(clazz);
-
-	if (!object || !class)
-		return false;
-
-	return vm_class_is_assignable_from(class, object->class);
-}
-
 static int
 native_vmfile_is_directory(struct vm_object *dirpath)
 {
@@ -727,6 +446,32 @@ static void native_vmthread_start(struct vm_object *vmthread, jlong stacksize)
 	vm_thread_start(vmthread);
 }
 
+static void native_vmthread_yield(void)
+{
+	vm_thread_yield();
+}
+
+static jint native_vmthread_interrupted(void)
+{
+	return vm_thread_interrupted(vm_thread_self());
+}
+
+static jint native_vmthread_isinterrupted(struct vm_object *vmthread)
+{
+	struct vm_thread *thread;
+
+	thread = (struct vm_thread *)field_get_object(vmthread, vm_java_lang_VMThread_vmdata);
+	return vm_thread_is_interrupted(thread);
+}
+
+static void native_vmthread_interrupt(struct vm_object *vmthread)
+{
+	struct vm_thread *thread;
+
+	thread = (struct vm_thread *)field_get_object(vmthread, vm_java_lang_VMThread_vmdata);
+	vm_thread_interrupt(thread);
+}
+
 static void native_vmobject_notify(struct vm_object *obj)
 {
 	vm_monitor_notify(&obj->monitor);
@@ -763,123 +508,84 @@ static void native_vmobject_wait(struct vm_object *object, jlong ms, jint ns)
 		vm_monitor_timed_wait(&object->monitor, ms, ns);
 }
 
-static struct vm_object *
-native_vmclassloader_findloadedclass(struct vm_object *classloader,
-				     struct vm_object *name)
-{
-	struct vm_class *vmc;
-	char *c_name;
-
-	c_name = vm_string_to_cstr(name);
-	if (!c_name)
-		return NULL;
-
-	vmc = classloader_find_class(c_name);
-	free(c_name);
-
-	if (!vmc)
-		return NULL;
-
-	if (vmc->classloader != classloader)
-		return NULL;
-
-	vm_class_ensure_init(vmc);
-	return vmc->object;
-}
-
-static struct vm_object *
-native_vmclassloader_loadclass(struct vm_object *name, jboolean resolve)
-{
-	/* XXX: what does @resolve parameter mean? */
-	struct vm_class *vmc;
-	char *c_name;
-
-	c_name = vm_string_to_cstr(name);
-	if (!c_name)
-		return NULL;
-
-	vmc = classloader_load(NULL, c_name);
-	free(c_name);
-	if (!vmc)
-		return NULL;
-
-	if (vm_class_ensure_object(vmc))
-		return NULL;
-
-	return vmc->object;
-}
-
-static jboolean
-native_vmclass_isinterface(struct vm_object *clazz)
-{
-	struct vm_class *class = vm_object_to_vm_class(clazz);
-	if (!class)
-		return false;
-
-	return vm_class_is_interface(class);
-}
-
 static struct vm_object *native_vmstring_intern(struct vm_object *str)
 {
 	return vm_string_intern(str);
 }
 
+static jint native_atomiclong_vm_supports_cs8(void)
+{
+	return false;
+}
+
 static struct vm_native natives[] = {
-	DEFINE_NATIVE("gnu/classpath/VMStackWalker", "getClassContext", &native_vmstackwalker_getclasscontext),
-	DEFINE_NATIVE("gnu/classpath/VMStackWalker", "getClassLoader", &native_vmclass_getclassloader),
-	DEFINE_NATIVE("gnu/classpath/VMSystemProperties", "preInit", &native_vmsystemproperties_preinit),
-	DEFINE_NATIVE("jato/internal/VM", "exit", &native_vmruntime_exit),
-	DEFINE_NATIVE("jato/internal/VM", "println", &native_vmruntime_println),
-	DEFINE_NATIVE("jato/internal/VM", "throwNullPointerException", &native_vm_throw_null_pointer_exception),
-	DEFINE_NATIVE("java/lang/VMClass", "getClassLoader", &native_vmclass_getclassloader),
-	DEFINE_NATIVE("java/lang/VMClass", "getDeclaredFields", &native_vmclass_get_declared_fields),
-	DEFINE_NATIVE("java/lang/VMClass", "getDeclaredConstructors", &native_vmclass_get_declared_constructors),
-	DEFINE_NATIVE("java/lang/VMClass", "getDeclaredMethods", &native_vmclass_get_declared_methods),
-	DEFINE_NATIVE("java/lang/VMClass", "getName", &native_vmclass_getname),
-	DEFINE_NATIVE("java/lang/VMClass", "forName", &native_vmclass_forname),
-	DEFINE_NATIVE("java/lang/VMClass", "isAnonymousClass", &native_vmclass_is_anonymous_class),
-	DEFINE_NATIVE("java/lang/VMClass", "isAssignableFrom", &native_vmclass_is_assignable_from),
-	DEFINE_NATIVE("java/lang/VMClass", "isArray", &native_vmclass_isarray),
-	DEFINE_NATIVE("java/lang/VMClass", "isInstance", &native_vmclass_isinstance),
-	DEFINE_NATIVE("java/lang/VMClass", "isInterface", &native_vmclass_isinterface),
-	DEFINE_NATIVE("java/lang/VMClass", "isPrimitive", &native_vmclass_isprimitive),
-	DEFINE_NATIVE("java/lang/VMClass", "getModifiers", &native_vmclass_getmodifiers),
-	DEFINE_NATIVE("java/lang/VMClass", "getComponentType", &native_vmclass_getcomponenttype),
-	DEFINE_NATIVE("java/lang/VMClass", "getInterfaces", &native_vmclass_get_interfaces),
-	DEFINE_NATIVE("java/lang/VMClass", "getSuperclass", &native_vmclass_get_superclass),
-	DEFINE_NATIVE("java/lang/VMClassLoader", "getPrimitiveClass", &native_vmclassloader_getprimitiveclass),
-	DEFINE_NATIVE("java/lang/VMClassLoader", "findLoadedClass", &native_vmclassloader_findloadedclass),
-	DEFINE_NATIVE("java/lang/VMClassLoader", "loadClass", &native_vmclassloader_loadclass),
-	DEFINE_NATIVE("java/io/VMFile", "isDirectory", &native_vmfile_is_directory),
-	DEFINE_NATIVE("java/lang/VMObject", "clone", &native_vmobject_clone),
-	DEFINE_NATIVE("java/lang/VMObject", "notify", &native_vmobject_notify),
-	DEFINE_NATIVE("java/lang/VMObject", "notifyAll", &native_vmobject_notify_all),
-	DEFINE_NATIVE("java/lang/VMObject", "getClass", &native_vmobject_getclass),
-	DEFINE_NATIVE("java/lang/VMObject", "wait", &native_vmobject_wait),
-	DEFINE_NATIVE("java/lang/VMRuntime", "exit", &native_vmruntime_exit),
-	DEFINE_NATIVE("java/lang/VMRuntime", "mapLibraryName", &native_vmruntime_maplibraryname),
-	DEFINE_NATIVE("java/lang/VMRuntime", "nativeLoad", &native_vmruntime_native_load),
-	DEFINE_NATIVE("java/lang/VMRuntime", "runFinalizationForExit", &native_vmruntime_run_finalization_for_exit),
-	DEFINE_NATIVE("java/lang/VMString", "intern", &native_vmstring_intern),
-	DEFINE_NATIVE("java/lang/VMSystem", "arraycopy", &native_vmsystem_arraycopy),
-	DEFINE_NATIVE("java/lang/VMSystem", "identityHashCode", &native_vmsystem_identityhashcode),
-	DEFINE_NATIVE("java/lang/VMSystem", "nanoTime", &native_vmsystem_nano_time),
-	DEFINE_NATIVE("java/lang/VMThread", "currentThread", &native_vmthread_current_thread),
-	DEFINE_NATIVE("java/lang/VMThread", "start", &native_vmthread_start),
-	DEFINE_NATIVE("java/lang/VMThrowable", "fillInStackTrace", &native_vmthrowable_fill_in_stack_trace),
-	DEFINE_NATIVE("java/lang/VMThrowable", "getStackTrace", &native_vmthrowable_get_stack_trace),
-	DEFINE_NATIVE("java/lang/reflect/Constructor", "getParameterTypes", &native_constructor_get_parameter_types),
-	DEFINE_NATIVE("java/lang/reflect/Constructor", "getModifiersInternal", &native_constructor_get_modifiers_internal),
-	DEFINE_NATIVE("java/lang/reflect/Constructor", "constructNative", &native_constructor_construct_native),
-	DEFINE_NATIVE("java/lang/reflect/Field", "get", &native_field_get),
+	DEFINE_NATIVE("gnu/classpath/VMStackWalker", "getClassContext", native_vmstackwalker_getclasscontext),
+	DEFINE_NATIVE("gnu/classpath/VMStackWalker", "getClassLoader", native_vmclass_getclassloader),
+	DEFINE_NATIVE("gnu/classpath/VMSystemProperties", "preInit", native_vmsystemproperties_preinit),
+	DEFINE_NATIVE("jato/internal/VM", "disableFault", native_vm_disable_fault),
+	DEFINE_NATIVE("jato/internal/VM", "enableFault", native_vm_enable_fault),
+	DEFINE_NATIVE("jato/internal/VM", "exit", native_vmruntime_exit),
+	DEFINE_NATIVE("jato/internal/VM", "println", native_vmruntime_println),
+	DEFINE_NATIVE("jato/internal/VM", "throwNullPointerException", native_vm_throw_null_pointer_exception),
+	DEFINE_NATIVE("java/io/VMFile", "isDirectory", native_vmfile_is_directory),
+	DEFINE_NATIVE("java/lang/VMClass", "forName", native_vmclass_forname),
+	DEFINE_NATIVE("java/lang/VMClass", "getClassLoader", native_vmclass_getclassloader),
+	DEFINE_NATIVE("java/lang/VMClass", "getComponentType", native_vmclass_getcomponenttype),
+	DEFINE_NATIVE("java/lang/VMClass", "getDeclaredConstructors", native_vmclass_get_declared_constructors),
+	DEFINE_NATIVE("java/lang/VMClass", "getDeclaredFields", native_vmclass_get_declared_fields),
+	DEFINE_NATIVE("java/lang/VMClass", "getDeclaredMethods", native_vmclass_get_declared_methods),
+	DEFINE_NATIVE("java/lang/VMClass", "getInterfaces", native_vmclass_get_interfaces),
+	DEFINE_NATIVE("java/lang/VMClass", "getModifiers", native_vmclass_getmodifiers),
+	DEFINE_NATIVE("java/lang/VMClass", "getName", native_vmclass_getname),
+	DEFINE_NATIVE("java/lang/VMClass", "getSuperclass", native_vmclass_get_superclass),
+	DEFINE_NATIVE("java/lang/VMClass", "isAnonymousClass", native_vmclass_is_anonymous_class),
+	DEFINE_NATIVE("java/lang/VMClass", "isArray", native_vmclass_isarray),
+	DEFINE_NATIVE("java/lang/VMClass", "isAssignableFrom", native_vmclass_is_assignable_from),
+	DEFINE_NATIVE("java/lang/VMClass", "isInstance", native_vmclass_isinstance),
+	DEFINE_NATIVE("java/lang/VMClass", "isInterface", native_vmclass_isinterface),
+	DEFINE_NATIVE("java/lang/VMClass", "isPrimitive", native_vmclass_isprimitive),
+	DEFINE_NATIVE("java/lang/VMClassLoader", "defineClass", native_vmclassloader_defineclass),
+	DEFINE_NATIVE("java/lang/VMClassLoader", "findLoadedClass", native_vmclassloader_findloadedclass),
+	DEFINE_NATIVE("java/lang/VMClassLoader", "getPrimitiveClass", native_vmclassloader_getprimitiveclass),
+	DEFINE_NATIVE("java/lang/VMClassLoader", "loadClass", native_vmclassloader_loadclass),
+	DEFINE_NATIVE("java/lang/VMObject", "clone", native_vmobject_clone),
+	DEFINE_NATIVE("java/lang/VMObject", "getClass", native_vmobject_getclass),
+	DEFINE_NATIVE("java/lang/VMObject", "notify", native_vmobject_notify),
+	DEFINE_NATIVE("java/lang/VMObject", "notifyAll", native_vmobject_notify_all),
+	DEFINE_NATIVE("java/lang/VMObject", "wait", native_vmobject_wait),
+	DEFINE_NATIVE("java/lang/VMRuntime", "availableProcessors", native_vmruntime_available_processors),
+	DEFINE_NATIVE("java/lang/VMRuntime", "exit", native_vmruntime_exit),
+	DEFINE_NATIVE("java/lang/VMRuntime", "gc", native_vmruntime_gc),
+	DEFINE_NATIVE("java/lang/VMRuntime", "mapLibraryName", native_vmruntime_maplibraryname),
+	DEFINE_NATIVE("java/lang/VMRuntime", "nativeLoad", native_vmruntime_native_load),
+	DEFINE_NATIVE("java/lang/VMRuntime", "runFinalizationForExit", native_vmruntime_run_finalization_for_exit),
+	DEFINE_NATIVE("java/lang/VMString", "intern", native_vmstring_intern),
+	DEFINE_NATIVE("java/lang/VMSystem", "arraycopy", native_vmsystem_arraycopy),
+	DEFINE_NATIVE("java/lang/VMSystem", "identityHashCode", native_vmsystem_identityhashcode),
+	DEFINE_NATIVE("java/lang/VMSystem", "nanoTime", native_vmsystem_nano_time),
+	DEFINE_NATIVE("java/lang/VMThread", "currentThread", native_vmthread_current_thread),
+	DEFINE_NATIVE("java/lang/VMThread", "interrupt", native_vmthread_interrupt),
+	DEFINE_NATIVE("java/lang/VMThread", "interrupted", native_vmthread_interrupted),
+	DEFINE_NATIVE("java/lang/VMThread", "isInterrupted", native_vmthread_isinterrupted),
+	DEFINE_NATIVE("java/lang/VMThread", "start", native_vmthread_start),
+	DEFINE_NATIVE("java/lang/VMThread", "yield", native_vmthread_yield),
+	DEFINE_NATIVE("java/lang/VMThrowable", "fillInStackTrace", native_vmthrowable_fill_in_stack_trace),
+	DEFINE_NATIVE("java/lang/VMThrowable", "getStackTrace", native_vmthrowable_get_stack_trace),
+	DEFINE_NATIVE("java/lang/reflect/Constructor", "constructNative", native_constructor_construct_native),
+	DEFINE_NATIVE("java/lang/reflect/Constructor", "getModifiersInternal", native_constructor_get_modifiers_internal),
+	DEFINE_NATIVE("java/lang/reflect/Constructor", "getParameterTypes", native_constructor_get_parameter_types),
+	DEFINE_NATIVE("java/lang/reflect/Field", "get", native_field_get),
+	DEFINE_NATIVE("java/lang/reflect/Field", "getLong", native_field_get_long),
 	DEFINE_NATIVE("java/lang/reflect/Field", "getModifiersInternal", native_field_get_modifiers_internal),
-	DEFINE_NATIVE("java/lang/reflect/Field", "set", &native_field_set),
-	DEFINE_NATIVE("java/lang/reflect/Field", "getType", &native_field_gettype),
-	DEFINE_NATIVE("java/lang/reflect/Method", "getParameterTypes", &native_method_get_parameter_types),
-	DEFINE_NATIVE("java/lang/reflect/Method", "invokeNative", &native_method_invokenative),
-	DEFINE_NATIVE("jato/internal/VM", "enableFault", &native_vm_enable_fault),
-	DEFINE_NATIVE("jato/internal/VM", "disableFault", &native_vm_disable_fault),
+	DEFINE_NATIVE("java/lang/reflect/Field", "getType", native_field_gettype),
+	DEFINE_NATIVE("java/lang/reflect/Field", "set", native_field_set),
+	DEFINE_NATIVE("java/lang/reflect/Method", "getModifiersInternal", native_method_get_modifiers_internal),
+	DEFINE_NATIVE("java/lang/reflect/Method", "getParameterTypes", native_method_get_parameter_types),
+	DEFINE_NATIVE("java/lang/reflect/Method", "getReturnType", native_method_getreturntype),
+	DEFINE_NATIVE("java/lang/reflect/Method", "invokeNative", native_method_invokenative),
+	DEFINE_NATIVE("java/util/concurrent/atomic/AtomicLong", "VMSupportsCS8", native_atomiclong_vm_supports_cs8),
 	DEFINE_NATIVE("sun/misc/Unsafe", "compareAndSwapInt", native_unsafe_compare_and_swap_int),
+	DEFINE_NATIVE("sun/misc/Unsafe", "compareAndSwapLong", native_unsafe_compare_and_swap_long),
 	DEFINE_NATIVE("sun/misc/Unsafe", "compareAndSwapObject", native_unsafe_compare_and_swap_object),
 	DEFINE_NATIVE("sun/misc/Unsafe", "objectFieldOffset", native_unsafe_object_field_offset),
 };
@@ -934,21 +640,19 @@ static void handle_jar(const char *arg)
 
 	jar_file = vm_jar_open(arg);
 	if (!jar_file) {
-		NOT_IMPLEMENTED;
+		fprintf(stdout, "Unable to open JAR file %s\n", arg);
 		exit(EXIT_FAILURE);
 	}
 
 	const char *main_class = vm_jar_get_main_class(jar_file);
 	if (!main_class) {
-		NOT_IMPLEMENTED;
+		fprintf(stdout, "Unable to look up main class in JAR file %s\n", arg);
 		exit(EXIT_FAILURE);
 	}
 
 	classname = strdup(main_class);
-	if (!classname) {
-		NOT_IMPLEMENTED;
-		exit(EXIT_FAILURE);
-	}
+	if (!classname)
+		die("out of memory");
 
 	/* XXX: Cheap solution. This can give funny results depending on where
 	 * you put the -jar relative to the -classpath(s). Besides, we should
@@ -956,9 +660,14 @@ static void handle_jar(const char *arg)
 	classloader_add_to_classpath(arg);
 }
 
-static void handle_gc(void)
+static void handle_nogc(void)
 {
-	gc_enabled = true;
+	dont_gc		= 1;
+}
+
+static void handle_maps(void)
+{
+	dump_maps = true;
 }
 
 static void handle_perf(void)
@@ -1010,11 +719,6 @@ static void handle_trace_exceptions(void)
 	opt_trace_exceptions = true;
 }
 
-static void handle_trace_gc_maps(void)
-{
-	opt_trace_gc_maps = true;
-}
-
 static void handle_trace_invoke(void)
 {
 	opt_trace_invoke = true;
@@ -1089,6 +793,16 @@ static void handle_verbose_gc(void)
 	verbose_gc = true;
 }
 
+static void handle_max_heap_size(const char *arg)
+{
+	max_heap_size = parse_long(arg);
+
+	if (!max_heap_size) {
+		fprintf(stderr, "%s: unparseable heap size '%s'\n", exe_name, arg);
+		usage(stderr, EXIT_FAILURE);
+	}
+}
+
 struct option {
 	const char *name;
 
@@ -1116,13 +830,15 @@ const struct option options[] = {
 	DEFINE_OPTION("help",		handle_help),
 
 	DEFINE_OPTION_ADJACENT_ARG("D",	handle_define),
+	DEFINE_OPTION_ADJACENT_ARG("Xmx", handle_max_heap_size),
 
 	DEFINE_OPTION_ARG("classpath",	handle_classpath),
 	DEFINE_OPTION_ARG("cp",		handle_classpath),
 	DEFINE_OPTION_ARG("jar",	handle_jar),
 	DEFINE_OPTION("verbose:gc",	handle_verbose_gc),
 
-	DEFINE_OPTION("Xgc",			handle_gc),
+	DEFINE_OPTION("Xnogc",			handle_nogc),
+	DEFINE_OPTION("Xmaps",			handle_maps),
 	DEFINE_OPTION("Xperf",			handle_perf),
 
 	DEFINE_OPTION_ARG("Xtrace:method",	handle_trace_method),
@@ -1133,7 +849,6 @@ const struct option options[] = {
 	DEFINE_OPTION("Xtrace:classloader",	handle_trace_classloader),
 	DEFINE_OPTION("Xtrace:compile",		handle_trace_compile),
 	DEFINE_OPTION("Xtrace:exceptions",	handle_trace_exceptions),
-	DEFINE_OPTION("Xtrace:gc-maps",		handle_trace_gc_maps),
 	DEFINE_OPTION("Xtrace:invoke",		handle_trace_invoke),
 	DEFINE_OPTION("Xtrace:invoke-verbose",	handle_trace_invoke_verbose),
 	DEFINE_OPTION("Xtrace:itable",		handle_trace_itable),
@@ -1166,8 +881,10 @@ static void parse_options(int argc, char *argv[])
 			break;
 
 		const struct option *opt = get_option(argv[optind] + 1);
-		if (!opt)
+		if (!opt) {
+			fprintf(stderr, "%s: unrecognized option '%s'\n", exe_name, argv[optind]);
 			usage(stderr, EXIT_FAILURE);
+		}
 
 		if (!opt->arg) {
 			opt->handler.func();
@@ -1217,6 +934,10 @@ do_main_class(void)
 	if (!loader)
 		return -1;
 
+	if (exception_occurred()) {
+		return -1;
+	}
+
 	struct vm_class *vmc = classloader_load(loader, classname);
 	if (!vmc) {
 		fprintf(stderr, "error: %s: could not load\n", classname);
@@ -1299,6 +1020,19 @@ static void gnu_classpath_autodiscovery(void)
 	}
 }
 
+static void print_proc_maps(void)
+{
+	char cmd[32];
+	pid_t self;
+
+	self = getpid();
+
+	snprintf(cmd, sizeof(cmd), "cat /proc/%d/maps", self);
+
+	if (system(cmd) != 0)
+		die("system");
+}
+
 int
 main(int argc, char *argv[])
 {
@@ -1323,12 +1057,18 @@ main(int argc, char *argv[])
 
 	parse_options(argc, argv);
 
+	if (dump_maps)
+		print_proc_maps();
+
 	classloader_init();
 
 	init_vm_objects();
+	init_vm_monitors();
 
 	jit_text_init();
 
+	gdb_init();
+
 	if (perf_enabled)
 		perf_map_open();
 
